/* GNU Chess 5.0 - atak.c - attack code
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   GNU Chess is based on the two research programs 
   Cobalt by Chua Kong-Sian and Gazebo by Stuart Cracraft.

   GNU Chess is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Chess is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info: 
     bug-gnu-chess@gnu.org
     cracraft@ai.mit.edu, cracraft@stanfordalumni.org, cracraft@earthlink.net
*/

#include <stdio.h>
#include <string.h>
#include "piece.h"
#include "util.h"
#include "bitboard.h"
#include "piece.h"
#include "move.h"
#include "magic.h"
#include "attack.h"


short AttackSq (board_t *board, short piece, short from, short to, short side)
/**************************************************************************
 *
 *  To determine if a piece on "from" from "side" attacks "to".
 *
 **************************************************************************/
{
   register BitBoard a;
   
   a = bitboards.BitPosArray[to];

   switch(piece){
   case knight:
       if (a & bitboards.MoveArray[knight][from]){
	   return true;
       }
       break;
   case king:
       if (a & bitboards.MoveArray[king][from]){
	   return true;
       }
       break;
   case pawn:
       if (a & bitboards.MoveArray[ptype[side]][from]){
	   return true;
       }
       break;
   case bishop:
       if(MagicBishopAttack(board,from) & a){
	   return true;
       }
       break;
   case rook:
       if(MagicRookAttack(board,from) & a){
	   return true;
       }
       break;
   case queen:
       if((MagicBishopAttack(board,from)|MagicRookAttack(board,from)) & a){
	   return true;
       }
       break;
   }

   return false;
}


short AttackToSq (board_t *board, short sq, short side)
/**************************************************************************
 *
 *  To determine if sq is attacked by any pieces from side.
 *
 **************************************************************************/
{
   register BitBoard *a;
   
   a = board->b[side];

   /* Knights */
   if (a[knight] & bitboards.MoveArray[knight][sq])
      return (true);

   /* Kings */
   if (a[king] & bitboards.MoveArray[king][sq])
      return (true);

   /* Pawns */
   if (a[pawn] & bitboards.MoveArray[ptype[1^side]][sq])
      return (true);
      
   /* Bishops & Queen */

   if(MagicBishopAttack(board,sq) & (a[bishop] | a[queen])){
       return true;
   }

   /* Rook e Queen */

   if(MagicRookAttack(board,sq) & (a[rook] | a[queen])){
       return true;
   }

   return (false);
}

short AttackToRegion(board_t *board, BitBoard b, short side)
/*********************************************************************
 * 
 * This returns true if one of the squares in b is attacked by side.
 * 
 *********************************************************************/
{
    int sq;
    while(b){
	sq=leadz(b);
	CLEARBIT(b,sq);
	if(AttackToSq (board, sq, side)){
	    return true;
	}
    }
    return false;
}


BitBoard AttackToSqBitBoard (board_t *board, int sq, int side)
/***************************************************************************
 *
 *  Generate a bitboard of all squares with pieces belonging to side
 *  which attack sq.
 *
 ***************************************************************************/
{
   register BitBoard *a, e;
   
   a = board->b[side];

   /* Knights */
   e = (a[knight] & bitboards.MoveArray[knight][sq]); 	

   /* Kings */
   e |= (a[king] & bitboards.MoveArray[king][sq]); 	

   /* Pawns */
   e |= (a[pawn] & bitboards.MoveArray[ptype[1^side]][sq]);
      

   /* Bishops & Queen */

   e |= MagicBishopAttack(board,sq) & (a[bishop] | a[queen]);

   /* Rook & Queen */

   e |= MagicRookAttack(board,sq) & (a[rook] | a[queen]);

   return (e);
}




BitBoard AttackXToSqBitBoard (board_t *board, int sq, int side)
/***************************************************************************
 *
 *  Generate a bitboard of all squares with pieces belonging to side
 *  which attack sq.  This routine is slightly different from 
 *  AttackToSqBitBoard
 *  as it includes X-ray attacks as well and these can go through the
 *  opponents pieces as well (e.g. a white R will attack the squares 
 *  beyond the black R). 
 *
 ***************************************************************************/
{
   register BitBoard *a, b, *c, *d, e, blocker;
   int t;
   
   a = board->b[side];
   d = board->b[1^side];

   /* Knights */
   e = (a[knight] & bitboards.MoveArray[knight][sq]); 	

   /* Kings */
   e |= (a[king] & bitboards.MoveArray[king][sq]); 	

   c = bitboards.FromToRay[sq];

   /* Bishops & Queen & Pawns */
   b = (a[pawn] & bitboards.MoveArray[ptype[1^side]][sq]);
   blocker = board->blocker;
   blocker &= ~(a[bishop] | a[queen] | d[bishop] | d[queen] | b);
   b |= (a[bishop] | a[queen]) & bitboards.MoveArray[bishop][sq];
   while (b){
      t = leadz (b);
      CLEARBIT (b, t); 
      if (!(c[t] & blocker & bitboards.NotBitPosArray[t]))
	 e |= bitboards.BitPosArray[t];
   }

   /* Rooks & Queen */
   b = (a[rook] | a[queen]) & bitboards.MoveArray[rook][sq];
   blocker = board->blocker;
   blocker &= ~(a[rook] | a[queen] | d[rook] | d[queen]);
   while (b) {
      t = leadz (b);
      CLEARBIT (b, t); 
      if (!(c[t] & blocker & bitboards.NotBitPosArray[t]))
	 e |= bitboards.BitPosArray[t];
   }

   return (e);
}


BitBoard AttackFromSqBitBoard (board_t *board, int sq, int side)
/***************************************************************************
 *
 *  Generate a bitboard of all squares attacked by the piece on sq. 
 *  side is the color of the piece. It is only important for pawns.
 *
 ***************************************************************************/
{
   enum Piece piece;
   piece = board->cboard[sq];
   switch (piece) {
      case pawn :
         return (bitboards.MoveArray[ptype[side]][sq]);
      case knight :
	 return (bitboards.MoveArray[knight][sq]);
      case bishop :
	return (MagicBishopAttack(board,sq));
      case rook :
	return (MagicRookAttack(board,sq));
      case queen :
	return (MagicQueenAttack(board,sq));
      case king :
	 return (bitboards.MoveArray[king][sq]);
      case empty:
	  return 0;
      case bpawn:
	  return 0;
   } 
   return (0);
}


BitBoard AttackXFromSqBitBoard (board_t *board, int sq, int side)
/***************************************************************************
 *
 *  Generate a bitboard of all squares attacked by a piece on sq.  
 *  side is the color. It is only important for pawns.
 *
 *  This routine is different from AttackFromSqBitBoard in that it includes 
 *  Xray attacks.
 *
 *  Caveat:  This routine does not take into account xrays through pawns.
 *
 ***************************************************************************/
{
   register BitBoard *a, b, blocker;
   int piece;

   a = board->b[side];
   piece = board->cboard[sq];
   blocker = board->blocker;
   b = 0;
   switch (piece){
      case pawn :
         b = bitboards.MoveArray[ptype[side]][sq];
         break;
      case knight :
	 b = bitboards.MoveArray[knight][sq];
         break;
      case bishop : /* falls through as queens move diagnonally */
      case queen :
	 blocker &= ~(a[bishop] | a[queen]);
	 b|=MagicBishopAttackEx(sq,blocker);
	 if (piece == bishop) /* Queen falls through as they move like rooks */
	    break;
         blocker = board->blocker;
      case rook :
	 blocker &= ~(a[rook] | a[queen]);
	 b|=MagicRookAttackEx(sq,blocker);
	 break;
      case king :
	 b = bitboards.MoveArray[king][sq];
         break;
   } 
   return (b);
}


void AttackGen (board_t *board, attack_t *attack)
/*************************************************************************
 *
 *  To generate the attack table.
 *  attack->b[side][0] holds the total attack tables.
 *  attack->b[side][pawn] holds the BitBoard of squares attacked by all pawns.
 *
 *************************************************************************/
{
   int side; 
   int sq;
   register BitBoard *a, b, *t, *a0;

   memset (attack, 0, sizeof (attack_t)); 
   for (side = white; side <= black; side++)
   {
      a = board->b[side];

      /* Knight */
      t = &(attack->b[side][knight]);
      b = a[knight];
      while (b){
         sq = leadz (b);
         CLEARBIT (b, sq);
         *t |= bitboards.MoveArray[knight][sq];
      }

      /* Bishops */
      t = &(attack->b[side][bishop]);
      b = a[bishop];
      while (b){
	 sq = leadz (b);
	 CLEARBIT (b, sq);
	 *t |= MagicBishopAttack(board,sq);
      }

      /*  Rooks */
      t = &(attack->b[side][rook]);
      b = a[rook];
      while (b){
	 sq = leadz (b);
	 CLEARBIT (b, sq);
	 *t |= MagicRookAttack(board,sq);
      }

      /*  Queen  */
      t = &(attack->b[side][queen]);
      b = a[queen];
      while (b){
	 sq = leadz (b);
	 CLEARBIT (b, sq);
	 *t |= MagicQueenAttack(board,sq);
      }

      /* King */
      t = &(attack->b[side][king]);
      b = a[king];
      while (b){
         sq = leadz (b);
         CLEARBIT (b, sq);
         *t |= bitboards.MoveArray[king][sq];
      }

      /*  pawns  */
      t = &(attack->b[side][pawn]);
      if (side == white){
         b = board->b[white][pawn] & ~bitboards.FileBit[0];
         *t |= (b >> 7);
         b = board->b[white][pawn] & ~bitboards.FileBit[7];
         *t |= (b >> 9);
      }
      else{
         b = board->b[black][pawn] & ~bitboards.FileBit[0];
         *t |= (b << 9);
         b = board->b[black][pawn] & ~bitboards.FileBit[7];
         *t |= (b << 7);
      }
      a0 = attack->b[side];
      a0[0] = a0[pawn] | a0[knight] | a0[bishop] | a0[rook] |
              a0[queen] | a0[king];
   }
}


int AttackIsEqual(attack_t *attack1, attack_t *attack2){
    int side;
    enum Piece piece;
    for(side=white;side<=black;side++){
	for(piece=empty;piece<=king;piece++){
	    if(attack1->b[side][piece]!=attack2->b[side][piece]){
		return false;
	    }
	}
    }
    return true;
}

int AttackValidate(board_t *board, attack_t *attack){
    attack_t attack1[1];
    AttackGen (board, attack1);
    if(!AttackIsEqual(attack,attack1)){
	return 1;
    }
    return 0;
}

