// board.c

#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "genmove.h"
#include "util.h"
#include "movelist.h"
#include "material.h"
#include "board.h"
#include "bitboard.h"
#include "move.h"
#include "util.h"
#include "piece.h"
#include "square.h"
#include "epd.h"
#include "attack.h"
#include "game.h"
#include "eval.h"
#include "log.h"
#include "pst.h"


static int OuterMostRook(BitBoard rook, int side, int wing, int king){
    /*********************************************************
     *  This function returns the outer most rook for 
     *  for side given the bitboard of rooks and the position
     *  of the king.
     *  
     *  The function does some sanity checking as well and
     *  returns -1 in the following cases
     *
     *  (1) The king is not on the first rank.
     *  (2) There is no rook on the correct side of the king.
     *
     *********************************************************/  
    
    BitBoard rooks_rank1;
    int outer_rook;
    if(RANK(king)!=rank1[side]){
	return -1;
    }
    rooks_rank1=rook & bitboards.RankBit[rank1[side]];
    if(rooks_rank1==NULLBITBOARD){
	return -1;
    }
    outer_rook=(wing==ks)?trailz(rooks_rank1):leadz(rooks_rank1);
    if(wing==ks && outer_rook<king){
	return -1;
    }
    if(wing==qs && outer_rook>king){
	return -1;
    }
    return outer_rook;

}


void BoardUpdateCastlingRooks(board_t *board){
    if((board->flag & WKINGCASTLE) && (board->wkrc==-1)){

	board->wkrc=OuterMostRook(board->b[white][rook],
				  white,
				  ks,
				  board->king[white]);
	if(board->wkrc==-1){
	    OutputConsole("Correcting invalid castling rights '%c'!\n",'K');
	    board->flag &= ~WKINGCASTLE;
	}
    }

    if((board->flag & WQUEENCASTLE) && (board->wqrc==-1)){
	board->wqrc=OuterMostRook(board->b[white][rook],
				  white,
				  qs,
				  board->king[white]);
	if(board->wqrc==-1){
	    OutputConsole("Correcting invalid castling rights '%c'!\n",'Q');
	    board->flag &= ~WQUEENCASTLE;
	}
    }
    if((board->flag & BKINGCASTLE) && (board->bkrc==-1)){
	board->bkrc=OuterMostRook(board->b[black][rook],
				  black,
				  ks,
				  board->king[black]);
	if(board->bkrc==-1){
	    OutputConsole("Correcting invalid castling rights '%c'!\n",'k');
	    board->flag &= ~BKINGCASTLE;
	}

    }
    if((board->flag & BQUEENCASTLE) && (board->bqrc==-1)){
	board->bqrc=OuterMostRook(board->b[black][rook],
				  black,
				  qs,
				  board->king[black]);
	if(board->bqrc==-1){
	    OutputConsole("Correcting invalid castling rights '%c'!\n",'q');
	    board->flag &= ~BQUEENCASTLE;
	}
    }
}


// Must be run after BoardUpdateMaterial

void BoardUpdatePST(board_t *board){
    BitBoard pieces;
    int color,piece;
    int t;
    board->PSTO=0;
    board->PSTE=0;
    for(color=white;color<=black;color++){
	for(piece=pawn;piece<=king;piece++){
	    pieces=board->b[color][piece];
	    while(pieces){
		t=leadz(pieces);
		CLEARBIT(pieces,t);
		board->PSTO+=P(piece,color,t,0);
		board->PSTE+=P(piece,color,t,1);
	    }
	}
    }
}

void BoardUpdateMaterial(board_t *board){
    enum Piece piece;
    int color;
    int count;
    board->O=0;
    board->E=0;
    for(color=white;color<=black;color++){
	board->material[color]=0;
	board->pmaterial[color]=0;
	for(piece=pawn;piece<=queen;piece++){
	    count=nbits(board->b[color][piece]);
	    board->material[color]+=count*evals.Value[piece];
	    board->O+=(color==white)?
		count*evals.ValueO[piece]:-count*evals.ValueO[piece];
	    board->E+=(color==white)?
		count*evals.ValueE[piece]:-count*evals.ValueE[piece];
	    if(piece!=pawn){
		board->pmaterial[color]+=count*evals.Value[piece];
	    }
	}
    }

}

void BoardUpdateMaterialIndex(board_t *board){
    int material[2][8];
    int color;
    enum Piece piece;
    for(color=white;color<=black;color++){
	for(piece=pawn;piece<=queen;piece++){
	    material[color][piece]=nbits(board->b[color][piece]);
	}
    }

    board->material_index=MaterialToIndex(material);

}


void BoardFromGame(board_t *board, game_t *game, int gamecnt){
  int i;
  int move;
  if(gamecnt==BOARD_LAST){
    gamecnt=game->RealGameCnt;
  }else if(gamecnt==BOARD_FIRST){
    gamecnt=game->InitialGameCnt;
  }
  BoardFromFen(board,game->InitialFen,game->variant==VARIANT_FRC);
  for(i=game->InitialGameCnt;i<gamecnt;i++){
    move=game->boards[i-game->InitialGameCnt].move;
    MoveMake(board,move);
  }
}

int BoardFromFen(board_t *board, const char *fen, int frc){
    int r,c,sq;
    const char *p;
    BitBoard rw,kw,rb,kb;
    int movenb,hmclock;
    int i;

    r = 56;
    c = 0;
    memset (board, 0, sizeof (board_t));

    p=fen;
    while (p && *p != ' '){
	sq = r + c;
	switch (*p)	{
	case 'P' :  SETBIT (board->b[white][pawn], sq);
	    board->material[white] += evals.ValueP;
	    board->O += evals.ValueO[pawn];
	    board->E += evals.ValueE[pawn];
	    break;	
	case 'N' :  SETBIT (board->b[white][knight], sq);
	    board->material[white] += evals.ValueN;
	    board->O += evals.ValueO[knight];
	    board->E += evals.ValueE[knight];
	    break;	
	case 'B' :  SETBIT (board->b[white][bishop], sq);
	    board->material[white] += evals.ValueB;
	    board->O += evals.ValueO[bishop];
	    board->E += evals.ValueE[bishop];
	    break;	
	case 'R' :  SETBIT (board->b[white][rook], sq);
	    board->material[white] += evals.ValueR;
	    board->O += evals.ValueO[rook];
	    board->E += evals.ValueE[rook];
	    break;	
	case 'Q' :  SETBIT (board->b[white][queen], sq);
	    board->material[white] += evals.ValueQ;
	    board->O += evals.ValueO[queen];
	    board->E += evals.ValueE[queen];
	    break;	
	case 'K' :  SETBIT (board->b[white][king], sq);
	    break;	
	case 'p' :  SETBIT (board->b[black][pawn], sq);
	    board->material[black] += evals.ValueP;
	    board->O -= evals.ValueO[pawn];
	    board->E -= evals.ValueE[pawn];
	    break;	
	case 'n' :  SETBIT (board->b[black][knight], sq);
	    board->material[black] += evals.ValueN;
	    board->O -= evals.ValueO[knight];
	    board->E -= evals.ValueE[knight];
	    break;	
	case 'b' :  SETBIT (board->b[black][bishop], sq);
	    board->material[black] += evals.ValueB;
	    board->O -= evals.ValueO[bishop];
	    board->E -= evals.ValueE[bishop];
	    break;	
	case 'r' :  SETBIT (board->b[black][rook], sq);
	    board->material[black] += evals.ValueR;
	    board->O -= evals.ValueO[rook];
	    board->E -= evals.ValueE[rook];
	    break;	
	case 'q' :  SETBIT (board->b[black][queen], sq);
	    board->material[black] += evals.ValueQ;
	    board->O -= evals.ValueO[queen];
	    board->E -= evals.ValueE[queen];
	    break;	
	case 'k' :  SETBIT (board->b[black][king], sq);
	    break;	
	case '/' :  r -= 8;
	    c = -1;
	    break;
	default  :  break;
	}
	if (isdigit (*p)){
	    c += (*p - '0');
	}else{
	    c++;
	}
	/* 
	 * Special case, a trailing "/" is accepted on the
	 * end of the board settings.
	 */
	
	if (r == -8 && p[1] == ' '){
	    r = 0;
	}
	
	if (r < 0 || c > 8){
	    Log("BoardFromFen(): Error 1\n");
	    return FEN_ERROR;
	}
	if (c == 8 && p[1] != '/' && p[1] != ' '){
	    Log("BoardFromFen(): Error 2\n");
	    return FEN_ERROR;
	}
	p++;
    
    }    
    board->pmaterial[white] = board->material[white] - 
	nbits(board->b[white][pawn]) * evals.ValueP;

    board->pmaterial[black] = board->material[black] - 
	nbits(board->b[black][pawn]) * evals.ValueP;


    board->king[white] = leadz (board->b[white][king]);
    board->king[black] = leadz (board->b[black][king]);
    BoardUpdateFriends (board); 
    BoardUpdateCBoard (board);
    BoardUpdateMvboard (board);
    BoardUpdateMaterialIndex(board);
    /*  Get side to move  */
    if (!++p){
	return FEN_ERROR;
    }
    if      (*p == 'w'){
	board->side = white; 
    }else if (*p == 'b') {
	board->side = black;
    }  else {
	Log("BoardFromFen(): Error 3\n");
	return FEN_ERROR;
    }
    
    /* Isn't this one cute? */
    if (!++p || *p != ' ' || !++p){
	Log("BoardFromFen(): Error 4\n");
	return FEN_ERROR;
    }
    /*  Castling status  */
    
    if(frc){
	board->flag |= FRCCASTLE;
    }

    rw = board->b[white][rook];
    kw = board->b[white][king];
    rb = board->b[black][rook];
    kb = board->b[black][king];
    board->wkrc=-1;
    board->wqrc=-1;
    board->bkrc=-1;
    board->bqrc=-1;
    while (p && *p != ' ') {
	if (false); 
	else if (*p == 'K'){
	    board->flag |= WKINGCASTLE;
	}else if (*p == 'Q'){
	    board->flag |= WQUEENCASTLE;
	}else if('A'<= *p && *p <='H'){
	    if(*p-'A'> FILE_(board->king[white])){
		board->flag |= WKINGCASTLE;
		board->wkrc=*p-'A';
	    }else{
		board->flag |= WQUEENCASTLE;
		board->wqrc=*p-'A';
	    }
	}else if (*p == 'k'){
	    board->flag |= BKINGCASTLE;
	}else if (*p == 'q'){
	    board->flag |= BQUEENCASTLE;
	}else if('a'<= *p && *p <='h'){
	    if(*p-'a'> FILE_(board->king[black])){
		board->flag |= BKINGCASTLE;
		board->bkrc=SQUARE(*p-'a',7);
	    }else{
		board->flag |= BQUEENCASTLE;
		board->bqrc=SQUARE(*p-'a',7);
	    }
	}else if (*p == '-') { p++; break; }
	else {
	    Log("BoardFromFen(): Error 5\n");
	    return FEN_ERROR;
	}
	p++;
    }
    
    BoardUpdateCastlingRooks(board);

    if (!p || *p != ' ' || !++p){
	Log("BoardFromFen(): Error 6\n");
	return FEN_ERROR;
    }
    

    /*
     * En passant square, can only be '-' or [a-h][36]
     * In fact, one could add more sanity checks here.
     */
    if (*p != '-') {
	if (!p[1] || *p < 'a' || *p > 'h' ||
	    !(p[1] == '3' || p[1] == '6')){
	    Log("BoardFromFen(): Error 7\n");
	    return FEN_ERROR;
	}
	board->ep = (*p - 'a') + (p[1] - '1')*8;
	p++;
    } else {
	board->ep = -1;
    }
    /* 
     *  Set ep passant square if necessary. 
     *  This deviates from the fen standard but
     *  many programs sin in this way.
     */
    
    if(board->ep!=-1 && !(board->b[board->side][pawn] & bitboards.MoveArray[ptype[1^board->side]][board->ep])){
	board->ep=-1;
    }
    board->in_check=AttackToSq (board,board->king[board->side], 1^board->side);
    board->xin_check=AttackToSq (board,board->king[1^board->side], board->side);
    
    
    p++;
    if(!(*p)|| (*p!=' ')){
	movenb=1;
	hmclock=0;
	goto epd_style;
	
    }
    p++;
    hmclock=0;
    movenb=0;
    // Read half move clock 
    if('0'<=*p && *p<='9'){   
	while(*p && *p!=' '){
	    if('0'<=*p && *p<='9'){
		hmclock*=10;
		hmclock+=*p-'0';
		p++;
	    }else{
		Log("BoardFromFen(): Error 9\n");
		return FEN_ERROR;
	    }
	}
	
	if(hmclock>=100){
	    hmclock=100;
	}
	// Read move number
	
	if(!(*p) || (*p!=' ') ){
	    Log("BoardFromFen(): Error 10\n");
	    return FEN_ERROR;
	}
	p++;
	while(*p && !isspace(*p)){
	    if('0'<=*p && *p<='9'){
		movenb*=10;
		movenb+=*p-'0';
		p++;
	    }else{
		Log("BoardFromFen(): Error 11\n");
		return FEN_ERROR;
	    }
	}   
    }
 epd_style:
    
    // Sometimes FEN's have move number zero which is strictly speaking illegal


    if(movenb==0) movenb=1;
    if(board->side==white){
	board->GameCnt=-2+2*movenb;
    }else{
	board->GameCnt=-1+2*movenb;
    }
    board->Game50=board->GameCnt-hmclock; 
    for(i=board->Game50;i<board->GameCnt;i++){
	board->stack[i-board->Game50]=0;
    }
    BoardUpdateHashKey (board);
    board->phase = Material[board->material_index].phase;
    BoardUpdatePST(board);

    ASSERT(BoardValidate(board)==0);

    return FEN_SUCCESS;
}

int BoardState(board_t *board){
  int ret=BOARD_PLAYING;
  movelist_t movelist[1];
  ASSERT(BoardValidate(board)==0);
  GenMoves (board,movelist);
  if (movelist->gencnt==0){
    if (board->in_check) {
       ret=(board->side==white?BOARD_BLACK_WINS:BOARD_WHITE_WINS);
     }else{
       ret=BOARD_STALEMATE;
     }
  }else if   ((board->GameCnt-board->Game50) >= 100){
    ret=BOARD_50MOVERULE;
  }else if(BoardRepeat(board) >= 2){
    ret=BOARD_3FOLD_REPETITION;
  }
  return ret;
}

void BoardToFen(board_t *board, char *fen){
  int r, c, sq, k;
  char c1;
  char tmp[MAXSTR];

  ASSERT(BoardValidate(board)==0);
  fen[0]=0;
  for (r = A8; r >= A1; r -= 8){
      k = 0;
      for (c = 0; c < 8; c++){
	  sq = r + c;
	  if (board->cboard[sq] == empty){
	    k++;
	  }else{
	    if (k){
	      sprintf (tmp, "%1d", k);
	      strcat(fen,tmp);
	    }
	    k = 0;
	    c1 = notation[board->cboard[sq]];
	    if (bitboards.BitPosArray[sq] & board->friends[black]){
	      c1 = tolower (c1);
	    }
	    sprintf (tmp, "%c", c1);
	    strcat(fen,tmp);
	  }
      }
      if (k){
	sprintf (tmp, "%1d", k);
	strcat(fen,tmp);
      }
      if (r > A1){
	sprintf (tmp, "/");
	strcat(fen,tmp);
      }
  }
 
  /* Other stuff */
  sprintf (tmp, (board->side == white ? " w " : " b "));
  strcat(fen,tmp);
  
  if (board->flag & WKINGCASTLE){
    sprintf (tmp, "K");
    strcat(fen,tmp);
  }
  if (board->flag & WQUEENCASTLE){
    sprintf (tmp, "Q");
    strcat(fen,tmp);
  }
  if (board->flag & BKINGCASTLE){
    sprintf (tmp, "k");
    strcat(fen,tmp);
  }
  if (board->flag & BQUEENCASTLE){
    sprintf (tmp, "q");
    strcat(fen,tmp);
  }
  if (!(board->flag & (WCASTLE | BCASTLE))){
    sprintf (tmp, "-");
    strcat(fen,tmp);
  }
  sprintf (tmp, " %s", (board->ep > -1 ? algbr[board->ep] : "-"));
  strcat(fen,tmp);
  sprintf (tmp," %d",board->GameCnt-board->Game50);
  strcat(fen,tmp);
  sprintf (tmp," %d",(board->GameCnt+2)/2);
  strcat(fen,tmp);
}

 


int BoardRepeat (board_t *board){
   int i, k;
   k = 0;
   for (i = board->GameCnt-4; i >= board->Game50; i-=2) 
   {
      if (board->stack[i-board->Game50] == board->HashKey)
         k++;
   }
   return (k);
}

int BoardRepeatEx (board_t *board, int ply)
/*****************************************************
 *
 * Look for repetions during the current search.
 *
 *****************************************************/
{
   int i, k, start, start1;
   k = 0;
   start=board->Game50;
   start1=board->GameCnt-ply;
   if(start1>=start){
       start=start1;
   }
   for (i = board->GameCnt-4; i >= start ; i-=2) 
   {
      if (board->stack[i-board->Game50] == board->HashKey)
         k++;
   }
   return (k);
}

int BoardIsEqual(board_t *first, board_t *second){
    int ret;
    ret=0;
    if(memcmp(&first->b,&second->b,sizeof(first->b))!=0){
	ret=-1;
    }else if(memcmp(&first->friends,&second->friends,sizeof(first->friends))!=0){
	ret=-2;
    }else if(memcmp(&first->material,&second->material,sizeof(first->material))!=0){
	ret=-3;
    }else if(memcmp(&first->pmaterial,&second->pmaterial,sizeof(first->pmaterial))!=0){
	ret=-4;
    }else if(memcmp(&first->O,&second->O,sizeof(first->O))!=0){
	ret=-10;
    }else if(memcmp(&first->E,&second->E,sizeof(first->E))!=0){
	ret=-12;
    }else if(memcmp(&first->castled,&second->castled,sizeof(first->castled))!=0){
	ret=-5;
    }else if(memcmp(&first->king,&second->king,sizeof(first->king))!=0){
	ret=-6;
    }else if(memcmp(&first->cboard,&second->cboard,sizeof(first->cboard))!=0){
	ret=-7;
    }else if(memcmp(&first->Mvboard,&second->Mvboard,sizeof(first->Mvboard))!=0){
	ret=-8;
    }else if(memcmp(&first->stack,&second->stack,sizeof(first->stack))!=0){
	ret=-9;
    }else if(first->ep!=second->ep){
	ret=1;
    }else  if(first->flag!=second->flag){
	ret=2;
    }else  if(first->blocker!=second->blocker){
	ret=3;
    }else  if(first->side!=second->side){
	ret=7;
    }else  if(first->Game50!=second->Game50){
	ret=8;
    }else  if(first->GameCnt!=second->GameCnt){
	ret=9;
    }else  if(first->HashKey!=second->HashKey){
	ret=10;
    }else  if(first->phase!=second->phase){ 
	ret=11;
    }else  if(first->PawnHashKey!=second->PawnHashKey){
	ret=12;
    }else if(first->in_check!=second->in_check){
	ret=13;
    }else if(first->xin_check!=second->xin_check){
	ret=14;
    }else if(first->wkrc!=second->wkrc){
	ret=15;
    }else if(first->wqrc!=second->wqrc){
	ret=16;
    }else if(first->bkrc!=second->bkrc){
	ret=17;
    }else if(first->bqrc!=second->bqrc){
	ret=18;
    }
    return ret;
}



short BoardValidate (board_t *board)
/***************************************************************************
 *
 *  Check the board to make sure that its valid.  Some things to check are
 *  a.  Both sides have only 1 king.
 *  b.  Side not on the move must not be in check.
 *  c.  If en passant square is set, check it is possible.
 *  d.  Check if castling status are all correct.
 *
 ***************************************************************************/
{
    int side, xside, sq, i;
    board_t board1[1];

    int ret=0;
    if (nbits (board->b[white][king]) != 1) {
	ret=1;
	goto fini;
    }
    if (nbits (board->b[black][king]) != 1) {
	ret=2;
	goto fini;
    }
    side = board->side;  
    xside = 1^side;

    if (board->ep > -1)
	{
	    sq = board->ep + (xside == white ? 8 : -8);
	    if (!(bitboards.BitPosArray[sq] & board->b[xside][pawn])){
		ret=4;
		goto fini;
	    }
	}

    if (!(bitboards.BitPosArray[board->king[white]] & board->b[white][king])){
	ret=5;
	goto fini;
    }
    if (!(bitboards.BitPosArray[board->king[black]] & board->b[black][king])){
	ret=9;
	goto fini;
    }
    if (board->flag & WKINGCASTLE){
	if (board->wkrc <board->king[white] || board->wkrc > H8){
	    ret=40;
	    goto fini;
	}
	if (!(bitboards.BitPosArray[board->wkrc] & board->b[white][rook])){
	    ret=6;
	    goto fini;
	}
    }
    if (board->flag & WQUEENCASTLE){
	if (board->wqrc <A1 || board->wqrc > board->king[white]){
	    ret=42;
	    goto fini;
	}
	if (!(bitboards.BitPosArray[board->wqrc] & board->b[white][rook])){
	    ret=8;
	    goto fini;;
	}
    }
    if (board->flag & BKINGCASTLE){
	if (board->bkrc < board->king[black]  || board->bkrc >H8){
	    ret=44;
	    goto fini;
	}
	if (!(bitboards.BitPosArray[board->bkrc] & board->b[black][rook])){
	    ret=10;
	    goto fini;
	}
    }
    if (board->flag & BQUEENCASTLE){
	if (board->bqrc <A8 || board->bqrc >board->king[black]){
	    ret=46;
	    goto fini;
	}
	if (!(bitboards.BitPosArray[board->bqrc] & board->b[black][rook])){
	    ret=12;
	    goto fini;
	}
    }
    
    if(nbits(board->b[white][pawn])>8){
	ret=14;
	goto fini;
    }
    if(nbits(board->b[black][pawn])>8){
	ret=15;
	goto fini;
    }
    if(board->GameCnt<board->Game50){
	ret=16;
	goto fini;
    }
    memcpy(board1,board,sizeof(board_t));
    BoardUpdateFriends(board1);
    BoardUpdateCBoard(board1);
    BoardUpdateHashKey(board1);
    BoardUpdateMaterialIndex(board1);
    BoardUpdateMaterial(board1);
    BoardUpdatePST(board1);
    for(i=0;i<64;i++){
	if(board->cboard[i]!=board1->cboard[i]){
	    ret=18;
	    goto fini;
	}
	if(board->Mvboard[i]<0){
	    ret=19;
	    goto fini;
	}
    }
    if(board->HashKey!=board1->HashKey){
	ret=20;
	goto fini;
    }
    if(board->friends[white]!=board1->friends[white]){
	ret=26;
	goto fini;
    }
    if(board->friends[black]!=board1->friends[black]){
	ret=27;
	goto fini;
    }
    if(board->blocker!=board1->blocker){
	ret=28;
	goto fini;
    }
    if(board->phase!=Material[board->material_index].phase){
	ret=29;
	goto fini;
    }
    if(board->in_check!=AttackToSq (board,board->king[board->side], 1^board->side)){
	ret=30;
	goto fini;
    }
    if(board->xin_check!=AttackToSq (board,board->king[1^board->side], board->side)){
	ret=31;
	goto fini;
    }
    if(board->material_index!=board1->material_index){
	ret=32;
	goto fini;
    }
    if(board->material_index!=board1->material_index){
	ret=32;
	goto fini;
    }
    if(board->material[white]!=board1->material[white]){
	ret=33;
	goto fini;
    }
    if(board->material[black]!=board1->material[black]){
	ret=34;
	goto fini;
    }
    if(board->pmaterial[white]!=board1->pmaterial[white]){
	ret=35;
	goto fini;
    }
    if(board->pmaterial[black]!=board1->pmaterial[black]){
	ret=36;
	goto fini;
    }
    if(board->O!=board1->O){
	ret=37;
	goto fini;
    }
    if(board->E!=board1->E){
	ret=39;
	goto fini;
    }
    if(board->PSTO!=board1->PSTO){
	ret=40;
	goto fini;
    }
    if(board->PSTE!=board1->PSTE){
	ret=41;
	goto fini;
    }
 fini:
    return ret;
}


void BoardShow (board_t *board)     
/*****************************************************************************
 *
 *  Display the board.  Not only that but display some useful information
 *  like whether enpassant is legal and castling state.
 *
 *****************************************************************************/ 
{
   int r, c, sq;
   int i;
   char fen[MAXFEN];

   BoardToFen(board,fen);
   for(i=0;i<board->GameCnt-board->Game50;i++){
     OutputConsole("stack[%d]="U64_FORMAT"\n",i,board->stack[i]);
   }
   OutputConsole("HashKey="U64_FORMAT"\n",board->HashKey);
   OutputConsole("Repeat=%d\n",BoardRepeat(board));
   OutputConsole("FEN=%s\n",fen);
   if (board->side == white){
     OutputConsole("white  ");
   }else{
     OutputConsole("black  ");
   }
   if (board->flag & WKINGCASTLE){
     OutputConsole("K");
   }
   if (board->flag & WQUEENCASTLE){
     OutputConsole("Q");
   }
   if (board->flag & BKINGCASTLE){
     OutputConsole("k");
   }
   if (board->flag & BQUEENCASTLE){
     OutputConsole("q");
   }
   if (board->ep > -1){
     OutputConsole("  %s", algbr[board->ep]);
   }
   OutputConsole("\n");
   OutputConsole("    +---+---+---+---+---+---+---+---+\n");
   for (r = 56; r >= 0; r -= 8){
       OutputConsole("  %d |",r/8+1);
      for (c = 0; c < 8; c++){
         sq = r + c;
         if (board->b[white][pawn]   & bitboards.BitPosArray[sq]){
	   OutputConsole(" P |");
         }else if (board->b[white][knight] & bitboards.BitPosArray[sq]){
	   OutputConsole(" N |");
         }else if (board->b[white][bishop] & bitboards.BitPosArray[sq]){
	   OutputConsole(" B |");
         }else if (board->b[white][rook]   & bitboards.BitPosArray[sq]){
	   OutputConsole(" R |");
         }else if (board->b[white][queen]  & bitboards.BitPosArray[sq]){
	   OutputConsole(" Q |");
	 }else if (board->b[white][king]   & bitboards.BitPosArray[sq]){
           OutputConsole(" K |");
         }else if (board->b[black][pawn]   & bitboards.BitPosArray[sq]){
           OutputConsole("*P*|");
         }else if (board->b[black][knight] & bitboards.BitPosArray[sq]){
           OutputConsole("*N*|");
         }else if (board->b[black][bishop] & bitboards.BitPosArray[sq]){
           OutputConsole("*B*|");
         }else if (board->b[black][rook]   & bitboards.BitPosArray[sq]){
           OutputConsole("*R*|");
         }else if (board->b[black][queen]  & bitboards.BitPosArray[sq]){
           OutputConsole("*Q*|");
         }else if (board->b[black][king]   & bitboards.BitPosArray[sq]){
           OutputConsole("*K*|");
         }else{
	     if((RANK(sq)+FILE_(sq))%2){
		 OutputConsole("   |");
	     }else{
		 OutputConsole(" . |");
	     }
	 }
      }
      OutputConsole("\n");
      OutputConsole("    +---+---+---+---+---+---+---+---+\n");
   }
   OutputConsole("      a   b   c   d   e   f   g   h\n");
}

void BoardShowEx(board_t *board){
    int r, c;
    
    BoardShow (board);
    
    for (r = 56; r >= 0; r -= 8) {
	for (c = 0; c < 8; c++){
	    OutputConsole("%2d ", board->Mvboard[r + c]);
	}
	OutputConsole("\n");
    }
    OutputConsole("\n");     
}

void BoardCopy(board_t *dst_board, board_t *src_board){
  memcpy(dst_board, src_board, sizeof(board_t));
}


void BoardUpdateHashKey (board_t *board)
/***************************************************************************
 *
 *  Calculates the hashkey for the current board position.  We sometimes
 *  need to do this especially when loading an EPD  position.
 *  Note:  The hashkey is a 64 bit unsigned integer number.  
 *  Added in pawnhashkey calculation.
 *
 ***************************************************************************/
{
   int sq, piece, color;
   BitBoard b;

   board->PawnHashKey = board->HashKey = (HashType) 0;
   for (color = white; color <= black; color++)
   {
      for (piece = pawn; piece <= king; piece++)
      {
	 b = board->b[color][piece];
	 while (b)
	 {
	    sq = leadz (b);
  	    CLEARBIT (b, sq);
	    board->HashKey ^= hashcode[color][piece][sq];
	    if (piece == pawn)
	       board->PawnHashKey ^= hashcode[color][piece][sq];
	 }
      }
   }

   /* Take into account castling status & en passant */
   if (board->ep > -1)
      board->HashKey ^= ephash[board->ep];
   if (board->flag & WKINGCASTLE)
      board->HashKey ^= WKCastlehash;
   if (board->flag & WQUEENCASTLE)
      board->HashKey ^= WQCastlehash;
   if (board->flag & BKINGCASTLE)
      board->HashKey ^= BKCastlehash;
   if (board->flag & BQUEENCASTLE)
      board->HashKey ^= BQCastlehash;

   if (board->side == white){
      board->HashKey ^= Sidehash;
   }
}

void BoardUpdateFriends (board_t *board)
/***************************************************************************
 *
 *  Update friend and enemy bitboard.
 *
 ***************************************************************************/
{
   register BitBoard *w, *b;

   w = board->b[white];
   b = board->b[black];
   board->friends[white] = 
      w[pawn] | w[knight] | w[bishop] | w[rook] | w[queen] | w[king];
   board->friends[black] = 
      b[pawn] | b[knight] | b[bishop] | b[rook] | b[queen] | b[king];
   board->blocker = board->friends[white] | board->friends[black];
}


void BoardUpdateCBoard (board_t *board)
/**************************************************************************
 *
 *  Updates cboard[].  cboard[i] returns the piece on square i.
 *
 **************************************************************************/
{
   BitBoard b;
   int piece, sq;

   memset (board->cboard, 0, sizeof (board->cboard));
   for (piece = pawn; piece <= king; piece++)
   {
      b = board->b[white][piece] | board->b[black][piece];
      while (b)
      {
         sq = leadz (b);
         CLEARBIT (b, sq);
         board->cboard[sq] = piece;
      }
   }
}


void BoardUpdateMvboard (board_t *board)
/**************************************************************************
 *
 *  Updates Mvboard[].  Mvboard[i] returns the number of times the piece
 *  on square i has moved.  
 *
 **************************************************************************/
{
   memset(board->Mvboard,0,sizeof(board->Mvboard));
}   




void BoardPut(board_t *board, int piece, int color, int sq){
    board->b[color][piece]|=bitboards.BitPosArray[sq];
    if(piece==king){
	board->king[color]=sq;
    }
    BoardUpdateFriends(board);
    BoardUpdateCBoard (board);
    BoardUpdateMvboard (board);
    BoardUpdateMaterialIndex(board);
    BoardUpdateHashKey (board);
    // TODO: deal with castling rights
}

void BoardInit(board_t *board){
    memset(board,0,sizeof(board_t));
    board->ep=-1;
    board->wqrc=-1;
    board->wkrc=-1;
    board->bqrc=-1;
    board->bkrc=-1;
}
