/* GNU Chess 5.0 - epd.c - EPD position definition code
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   GNU Chess is based on the two research programs 
   Cobalt by Chua Kong-Sian and Gazebo by Stuart Cracraft.

   GNU Chess is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Chess is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info: 
     bug-gnu-chess@gnu.org
*/
/*
 *
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "util.h"
#include "eval.h"
#include "bitboard.h"
#include "material.h"
#include "piece.h"
#include "square.h"
#include "move.h"
#include "epd.h"
#include "log.h"
#include "board.h"
#include "attack.h"
#include "protocol.h"


#define EPDCLOSE 1

epd_iter_t EPDIter[1];

void EPDIterInit(epd_iter_t *epd_iter){
  epd_iter->fp=NULL;
}

int EPDIterStart(epd_iter_t *epd_iter, char *file){
  if(epd_iter->fp!=NULL){
    fclose(epd_iter->fp);
  }
  epd_iter->fp = fopen (file, "r");
  if (epd_iter->fp == NULL) {
      return false;
  }
  return true;
}

void EPDIterClose(epd_iter_t *epd_iter){
  if(!epd_iter->fp){
    return;
  }
  fclose (epd_iter->fp);
  epd_iter->fp=NULL;
}

int EPDIterNext(epd_iter_t *epd_iter, epd_t *epd){
    char *dummy;
    char line[MAXSTR];
    int ret;
    if(epd_iter->fp==NULL){
	return false;
    }
 next_line:
    dummy=fgets (line, MAXSTR-1, epd_iter->fp);
    if (!feof(epd_iter->fp)) {
	ret=EPDFromString(epd,line);

	if (ret != EPD_SUCCESS) {
	    goto next_line;
	}
    }else{ 
	fclose (epd_iter->fp);
	epd_iter->fp = NULL;
	return false;
    }
    return true;
}

int EPDFieldFromString(char *dst, const char *src, const char *name){
    // TODO: make this more robust by not using strstr.
    const char *str_p;
    char *str_p1;
    const char *str_p2;
    char dummy[MAXEPDFIELD];
    char tmp[MAXEPDFIELD];
    int found;
    int in_string;
    dst[0]='\0';
    found=false;
    str_p=src;
    while(!found){
	str_p = strstr(str_p, name);
	if (str_p) {
	    // Dirty. Check that we are not in a string
	    // by counting the number of quotes to the left.
	    in_string=false;
	    for(str_p2=src;str_p2<str_p;str_p2++){
		if((*str_p2)=='\"'){
		    in_string=!in_string;
		}
	    }
	    if(in_string){
		str_p++;
		continue;
	    }else{
		sscanf (str_p, "%240s %63[^;];", dummy, tmp);
		found=true;
	    }
	}
	break;
    }
    if(!found){
	    return EPD_ERROR;
    }
    // eliminate quotes
    if((str_p=strchr(tmp,'\"')) && (str_p1=strchr(str_p+1,'\"'))){
	*str_p1='\0';
	strcpy(dst,str_p+1);
    }else{
	strcpy(dst,tmp);
    }
    return EPD_SUCCESS;
}

/*
 * Returns EPD_SUCCESS on success, EPD_ERROR on error. We try to be
 * quite tough on the format. However, as of yet no legality checking
 * is done and the board is not reset on error, this should be done by
 * the caller.
 */

int EPDFromString (epd_t *epd, const char *p)
/**************************************************************************
 *   
 *  Parses an EPD input line.  A few global variables are updated e.g.
 *  current board, side to move, en passant, castling status, etc.
 *
 **************************************************************************/
{
    int fc;
    char *f;
    char bm[MAXEPDFIELD];
    char fen[MAXFEN];
    
    epd->id[0] = '\0';

    f=fen;
   
    /* Try to read first 4 fields. These represent the position. */

    fc=0;
    while(fc<4 && *p!='\0'){
	if((*(f++)=*(p++))==' ' || *p=='\0' || *p=='\r' || *p=='\n'){
	    fc++;
	}
    }
    *f='\0';
    if(fc<4){
       return EPD_ERROR;
    }
    if(BoardFromFen(epd->board,fen,Protocol->game->variant==VARIANT_FRC)!=FEN_SUCCESS){
	return EPD_ERROR;
    }

    /* The opcodes are optional, so we should not generate errors here */
    
    /*  Read in best move; "bm" operator */

    if(EPDFieldFromString(bm,p,"bm")==EPD_SUCCESS){
	MoveListFromString(epd->board,epd->bm,bm);
    }
    
    EPDFieldFromString(epd->id,p,"id");
    EPDFieldFromString(epd->c0,p,"c0");
    
    return EPD_SUCCESS;
}

void EPDShow(epd_t *epd){
    BoardShow(epd->board);
    OutputConsole("id: %s\n",epd->id); //bm=%s\nmovelist:",epd->id,SANmv);
    OutputConsole("bm: ");
    MoveListShow(epd->board,epd->bm);
    if(epd->c0[0]!='\0'){
	OutputConsole("c0: %s\n",epd->c0);
    }
}

