/* GNU Chess 5.0 - eval.h - evaluation symbolic definitions
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   GNU Chess is based on the two research programs 
   Cobalt by Chua Kong-Sian and Gazebo by Stuart Cracraft.

   GNU Chess is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Chess is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info: 
     bug-gnu-chess@gnu.org
     cracraft@ai.mit.edu, cracraft@stanfordalumni.org, cracraft@earthlink.net
*/

#ifndef EVAL_H
#define EVAL_H

// includes

#include "bitboard.h"
#include "searchdata.h"
#include "searchstats.h"
#include "option.h"
#include "ptable.h"
#include "attack.h"

// macros

#define OPENING(board)	     ((board)->phase <= 2)
#define ENDING(board)       ((board)->phase >= 6)

// XXX make this into an array

#define BACKWARDPAWN(board)   -(8+(board)->phase)
#define PHASEMIX(board,op,eg) (((255-Material[(board)->material_index].phase256)*((int)(op))+Material[(board)->material_index].phase256*((int)(eg)))/256)

// types

typedef struct {
/***************************************************************************
 *
 * General
 *
 ***************************************************************************/
    short stm_bonusO;
    short stm_bonusE;
/***************************************************************************
 *
 * Generic Material values (should be deprecated)
 *
 ***************************************************************************/
    short ValueP;
    short ValueN;
    short ValueB;
    short ValueR;
    short ValueQ;
    short ValueK;
    short Value[7];
/****************************************************************************
 *
 * Material values
 *
 ****************************************************************************/

    short ValueO[7];
    short ValueE[7];

/****************************************************************************
 *
 *  The various evaluation constants.
 *
 ****************************************************************************/
    short tradepiece;
    short tradepawns;
    short hungpenalty;
    short rookmoved;
    short kingmoved;
    short notcastled;
    short pfactor;
    short earlyminorrepeat;
    short undevelopedminor;
/****************************************************************************
 *
 *  Pawns  
 *
 ***************************************************************************/
    short eight_pawns;
    short stonewall;
    short lockedpawns;
    short earlywingpawnmove;
    short earlycenterrepeat;
    short pawnbaseattack;
    short blockedpawn;
    short connectedpp;
    short pawnnearking;
    //    short attackweakpawn;
    short fianchetto_target;
    short doubled_pawn;
    short isolated_pawn_open;
    short isolated_doubled_pawn_open;
    short isolated_pawn_normal;
    short isolated_doubled_pawn_normal;
    short pawn_endgame_bonus;
    short passer_seedO;
    short passer_seedE;
    short passerO[2][8];
    short passerE[2][8];
    short free_passer_seedO;
    short free_passer_seedE;
    short free_passerO[2][8];
    short free_passerE[2][8];
/****************************************************************************
 *
 *  Knights  
 *
 ***************************************************************************/
    short knightonrim;
    short outpostknight;
    short pinnedknight;
    short knighttrapped;
/****************************************************************************
 *
 *  Bishops
 *
 ***************************************************************************/

    short doubledbishops;
    short outpostbishop;
    short fianchetto;
    short goodendingbishop;
    short bishoptrapped;
    short pinnedbishop;

/****************************************************************************
 *
 *  Rooks
 *
 ***************************************************************************/
    short rook7rank;
    short rooks7rank;
    short rookhalffile;
    short rookopenfile;
    short rookbehindpp;
    short rookinfrontpp;
    short pinnedrook;
    short rooktrapped;
    short rookliberated;
/****************************************************************************
 *
 *  Queen
 *
 ***************************************************************************/
    short earlyqueenmove;
    short queennearking;
    short pinnedqueen;    
    short queen_not_present;
/****************************************************************************
 *
 *  King
 *
 ***************************************************************************/
    short kingopenfile;
    short kingopenfile1;
    short attackking;
    short defendking;
    short hopen;
    short gopen;
    short king_defender_deficit;
    short king_back_rank_weak;
    short rupture;
    
/***************************************************************************
 *
 * Mobility
 *
 ***************************************************************************/

    short defensefriendlyking;
    short mobility;
    short attackenemypiece;

    short mobilityO[7];
    short mobilityE[7];
    short attackenemyking[7];
    short centercontrol[7];

    short attackweakpawnKO;
    short attackweakpawnKE;
    short attackweakpawnQ;
    short attackweakpawnR;
    short attackweakpawnB;
    short attackweakpawnN;



} eval_t;

typedef struct {
    /* Filled (for side) by ScoreP */
    BitBoard passed[2];
    BitBoard weak[2];
    BitBoard attacked[2];
    BitBoard backward[2];
    BitBoard blocked[2];
    int real_pawns[2];
    /* Filled (for side) by the various Score* routines */
    int score_p[2];
    int score_b[2];
    int score_n[2];
    int score_r[2];
    int score_q[2];
    int score_dev[2];
    /* ScoreK1 (lazy eval) */
    int score_k_1[2];
    /* Filled by PinsFind */
    BitBoard pinned;
    /* Collected when invoking CTL for the various pieces */
    int score_k_attack[2];
    int score_k_attack_nb[2];
    /* ScoreK (full eval) */
    int score_ks[2];
    /* ScoreK1, possibly updated in ScoreK */
    int score_k[2];
    /* Filled by CTL */
    int ctl[2][7];
    attack_t attack;

} evaldata_t;


// functions

void EvalInit(option_list_t *options);

int ScoreP (board_t *board, 
	    short side,
	    evaldata_t *evaldata);

int ScoreN (board_t *board, 
	    short side,
	    evaldata_t *evaldata);

int ScoreB (board_t *board, 
	    short side,
	    evaldata_t *evaldata);

int ScoreR (board_t *board, 
	    short side,
	    evaldata_t *evaldata);

int ScoreQ (board_t *board, 
	    short side,
	    evaldata_t *evaldata);

int ScoreK (board_t *board, 
	    short side,
	    evaldata_t *evaldata);

int ScoreDev (board_t *board, 
	      short side,
	      evaldata_t *evaldata);

int Evaluate (board_t *board, 
	      searchdata_t *searchdata, 
	      searchstats_t *searchstats, 
	      evaldata_t *evaldata,
	      int alpha, 
	      int beta);

short EvaluateDraw (board_t *board);

void EvalDataInit(evaldata_t *evaldata);

// variables

eval_t evals;

#endif /* !EVAL_H */
