#include "pin.h"
#include "piece.h"
#include "log.h"

int PinnedOnKing (board_t *board, int sq, int side)
/***************************************************************************
 *
 *  Determine if the piece on sq is pinned against the King.
 *  Side is the color of the piece.  
 *  Caveat: PinnedOnKing should only be called by GenCheckEscapes().
 *  The more generic FindPins() function should be used for evaluating
 *  pins against other pieces.
 *
 ***************************************************************************/
{
   int xside;
   int KingSq, dir, sq1;
   BitBoard b, blocker;

   KingSq = board->king[side];
   if ((dir = directions[KingSq][sq]) == -1)
      return (false);

   xside = 1 ^ side;
   blocker = board->blocker;
 
   /*  Path from piece to king is blocked, so no pin */
   if (bitboards.FromToRay[KingSq][sq] & bitboards.NotBitPosArray[sq] & blocker)
      return (false);
   b = (bitboards.Ray[KingSq][dir] ^ bitboards.FromToRay[KingSq][sq]) & blocker;
   if (b == NULLBITBOARD)
      return (false);
   sq1 = (sq > KingSq ? leadz (b) : trailz (b));

   /*  If diagonal  */
   if (dir <= 3 && 
	bitboards.BitPosArray[sq1] & 
       (board->b[xside][queen] | board->b[xside][bishop]))
      return (true);
   
   /*  Rank / file  */  
   if (dir >= 4 && 
	bitboards.BitPosArray[sq1] & (board->b[xside][queen] | board->b[xside][rook]))
      return (true);

   return (false);
}



BitBoard PinsFind (board_t *board, attack_t *attack)
/***************************************************************************
 *
 *  This function creates a bitboard of all pieces which are pinned
 *  (for both sides).
 *
 ***************************************************************************/
{
   int side, xside;
   int sq, sq1;
   BitBoard b, c, e, f, t, *p, pin;

   ASSERT(AttackValidate(board,attack)==0);
   
   pin = NULLBITBOARD;
   t = board->friends[white] | board->friends[black];
   for (side = white; side <= black; side++)
   {
      xside = 1^side;
      p = board->b[xside];

      /*  Check if bishop is pinning anything */
      e = p[rook] | p[queen] | p[king];
      e |= (p[bishop] | p[knight]) & ~(attack->b[xside][0]);
      b = board->b[side][bishop];
      while (b)
      {
         sq = leadz (b);
         CLEARBIT (b, sq); 

	 c = bitboards.MoveArray[bishop][sq] & e;
         while (c)
         {
	    sq1 = leadz (c);
	    CLEARBIT (c, sq1);
	    f = t & bitboards.NotBitPosArray[sq] & bitboards.FromToRay[sq1][sq];
            if ((board->friends[xside] & f) && nbits (f) == 1)
	       pin |= f;
         }
      }

      /*  Check if rook is pinning anything */
      e = p[queen] | p[king];
      e |= (p[rook] | p[bishop] | p[knight]) & ~(attack->b[xside][0]);
      b = board->b[side][rook];
      while (b)
      {
         sq = leadz (b);
         CLEARBIT (b, sq); 

	 c = bitboards.MoveArray[rook][sq] & e;
         while (c)
         {
	    sq1 = leadz (c);
	    CLEARBIT (c, sq1);
	    f = t & bitboards.NotBitPosArray[sq] & bitboards.FromToRay[sq1][sq];
            if ((board->friends[xside] & f) && nbits (f) == 1)
	       pin |= f;
         }
      }

      /*  Check if queen is pinning anything */
      e = board->b[xside][king];
      e |= (p[queen] | p[rook] | p[bishop] | p[knight]) & ~(attack->b[xside][0]);
      b = board->b[side][queen];
      while (b)
      {
         sq = leadz (b);
         CLEARBIT (b, sq); 

	 c = bitboards.MoveArray[queen][sq] & e;
         while (c)
         {
	    sq1 = leadz (c);
	    CLEARBIT (c, sq1);
	    f = t & bitboards.NotBitPosArray[sq] & bitboards.FromToRay[sq1][sq];
            if ((board->friends[xside] & f) && nbits (f) == 1)

	       pin |= f;
         }
      }
   }
      
   return pin;
}


