/* GNU Chess 5.0 - quiesce.c - quiescence search code
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   GNU Chess is based on the two research programs 
   Cobalt by Chua Kong-Sian and Gazebo by Stuart Cracraft.

   GNU Chess is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Chess is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info: 
   bug-gnu-chess@gnu.org
   cracraft@ai.mit.edu, cracraft@stanfordalumni.org, cracraft@earthlink.net
*/
/*
 *
 */

#include <stdio.h>
#include "util.h"
#include "move.h"
#include "ttable.h"
#include "eval.h"
#include "search.h"
#include "movelist.h"
#include "searchstats.h"
#include "searchdata.h"
#include "sort.h"
#include "swap.h"
#include "genmove.h"
#include "log.h"
#include "piece.h"

int Quiesce (board_t *board, 
	     searchdata_t *searchdata, 
	     searchstats_t *searchstats, 
	     uint8_t ply, 
	     int alpha, 
	     int beta,
	     int depth,
	     int *best_move_out)
/**************************************************************************
 *
 *  Our quiescent search.  This quiescent search is able to recognize
 *  mates.
 *
 **************************************************************************/
{
    uint8_t side, xside;
    int best,  score, savealpha, pat;
    leaf *p, *pbest;
    undo_t undo[1];
    int leafnode=true;
    int skip_stalemate_check=false;
    int best_move;
    int ret;
    int movecount=0;
    pickstate_t pickstate[1];
    int in_check;
    evaldata_t evaldata[1];
    int swapoff;
    int forward_check=false;

    best_move=NOMOVE;
    *best_move_out=NOMOVE;

    searchstats->QuiesCnt++;

    if(ply>=searchdata->seldepth){
	searchdata->seldepth=ply;
    }


    // emergency exit

    if(ply>=MAXPLYDEPTH){
	// evaldata is initialized in Evaluate
	return board->in_check?alpha:Evaluate(board, 
					      searchdata, 
					      searchstats, 
					      evaldata,
					      alpha,
					      beta);
    }



    if (EvaluateDraw (board)){
	return (DRAWSCORE);
    }

    side = board->side;
    xside = 1^side;
    pat=best=-INFIN;
    in_check = board->in_check;
    PickStateInit(pickstate);

    savealpha = alpha;
    pbest = NULL;

    if(!board->in_check){
	int buffer;
	//	if(depth==0){
	//	    buffer=search.SafetyDelta;
	//	}else{
	//	    buffer=0;
	//	}
	// evaldata is initialized in Evaluate
	pat=best = Evaluate (board, 
			     searchdata, 
			     searchstats, 
			     evaldata,
			     alpha/*-buffer*/, 
			     beta/*+buffer*/);
    }
    if (best >= beta){
	goto done;
    }

    alpha = MAX(best, alpha);

    while(true){
	
	if(board->in_check){
	    ret=PhasePick1(board,searchdata,pickstate,&p,ply,depth);
	}else{
	    ret=PhasePickQuiesce(board,searchdata,pickstate,&p,ply,depth);
	}
	 
	if(!ret){
	    break;
	}

	int to_value;
	int from_value;
	to_value=evals.Value[board->cboard[TOSQ(p->move)]];
	from_value=evals.Value[board->cboard[FROMSQ(p->move)]];

	if(!in_check && ((p->move & PROMOTION) && (PROMOTEPIECE(p->move)!=queen))){
	    continue;
	}
	forward_check=false;
	if (beta==alpha+1 /*not a PV node */
	    && !in_check  /* not check evasion */
	    && !(p->move & (PROMOTION|ENPASSANT))      /* not a promotion  or enpassant */
	    && pickstate->phase!=PICKQUIETCHECKS){       /* not a check */
	    ASSERT(to_value!=0);

	    forward_check=MoveForwardCheck(board,p->move);
	    if(!forward_check){
		if(pat+to_value< alpha-search.DeltaMargin){
		    best=MAX(best,pat+to_value+search.DeltaMargin);
		    continue;
		}
	    }
	}

	if(!in_check  && from_value>to_value){
	    
	    if(SwapOffSignEx(board,p->move,0,false)<0){
		continue;
	    }
	
	}


	MoveMakeS (board, &p->move, undo);
	ASSERT(!forward_check || board->in_check);
	if (board->xin_check){
	    MoveUnMakeS (board, undo);
	    continue;
	}


	movecount++;
	score = -Quiesce (board, 
			  searchdata, 
			  searchstats, 
			  ply+1, 
			  -beta, 
			  -alpha,
			  depth-1,
			  &best_move);
	leafnode=false;
	MoveUnMakeS (board, undo);
	if (score > best) {
	    best = score;
	    pbest = p;
	    *best_move_out=p->move;
	    if (best >= beta){
		goto done;
	    }
	    alpha = MAX (alpha, best);
	}
    }

    if(board->in_check && movecount==0){
	return (-MATE+(ply+1)-2);
    }
 done:


    if(DRAWSCORE<=best && best<=alpha){
	skip_stalemate_check=true;
    }
    if(DRAWSCORE>=best && best>=beta){
	skip_stalemate_check=true;
    }
    if(!skip_stalemate_check && leafnode && board->pmaterial[side] <= evals.ValueR && GenStalemateCheck(board)){
    	return DRAWSCORE;
    }
    ASSERT(best>=-MATE-1);
    ASSERT(best<=MATE+1);


    return (best);
}
 
