/* GNU Chess 5.0 - swap.c - static exchange evaluator code
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   GNU Chess is based on the two research programs 
   Cobalt by Chua Kong-Sian and Gazebo by Stuart Cracraft.

   GNU Chess is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Chess is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info: 
     bug-gnu-chess@gnu.org
     cracraft@ai.mit.edu, cracraft@stanfordalumni.org, cracraft@earthlink.net
*/

#include <stdio.h>
#include <string.h>

#include "eval.h"
#include "move.h"
#include "piece.h"
#include "swap.h"
#include "log.h"

static const int xray[7] = { 0, 1, 0, 1, 1, 1, 0 };


int SwapOffEx (board_t *board, int move, int move_already_made)
/****************************************************************************
 *
 *  A Static Exchange Evaluator (or SEE for short).
 *  First determine the target square.  Create a bitboard of all squares
 *  attacking the target square for both sides.  Using these 2 bitboards,
 *  we take turn making captures from smallest piece to largest piece.
 *  When a sliding piece makes a capture, we check behind it to see if
 *  another attacker piece has been exposed.  If so, add this to the bitboard
 *  as well.  When performing the "captures", we stop if one side is ahead
 *  and doesn't need to capture, a form of pseudo-minimaxing.
 *
 ****************************************************************************/
{
    int f, t, piece, lastval;
    int side_attack_from_sq, xside_attack_from_sq;
    int side, xside; 
    int swaplist[MAXPLYDEPTH], n;
    BitBoard side_attack, xside_attack, xside_attack_from_piece, side_attack_from_piece;
    
    f = FROMSQ (move);
    t = TOSQ (move);
    if(!move_already_made){
	side = ((board->friends[white] & bitboards.BitPosArray[f]) ? white : black);
    }else{
	side = ((board->friends[white] & bitboards.BitPosArray[t]) ? white : black);
    }
    xside = 1^side;
    
    /*  Squares attacking t for side and xside  */
    side_attack = AttackToSqBitBoard (board, t, side);
    xside_attack = AttackToSqBitBoard (board, t, xside);
    if(!move_already_made){
	CLEARBIT(side_attack, f);
	if (xray[board->cboard[f]]){
	    AddXrayPiece (board,t, f, side, &side_attack, &xside_attack);
	}
	if (move & PROMOTION){
	    swaplist[0] = evals.Value[PROMOTEPIECE (move)] - evals.ValueP;
	    lastval = -evals.Value[PROMOTEPIECE (move)];
	} else {
	    swaplist[0] = (move & ENPASSANT ? evals.ValueP : evals.Value[board->cboard[t]]);
	    lastval = -evals.Value[board->cboard[f]];
	}
    }else{
	lastval=-evals.Value[board->cboard[t]];
	swaplist[0]=0; /* we cannot know what piece was there */
    }
    n = 1;
    while (1){
	if (xside_attack == NULLBITBOARD){
	    break;
	}
	for (piece = pawn; piece <= king; piece++) {
	    xside_attack_from_piece = xside_attack & board->b[xside][piece]; 
	    if (xside_attack_from_piece){
		xside_attack_from_sq = leadz (xside_attack_from_piece);
		CLEARBIT (xside_attack, xside_attack_from_sq);
		if (xray[piece]){
		    AddXrayPiece (board, t, xside_attack_from_sq, xside, &xside_attack, &side_attack);
		}
		swaplist[n] = swaplist[n-1] + lastval;
		n++;
		lastval = evals.Value[piece];
		break;
	    }
	}
	
	if (side_attack == NULLBITBOARD){
	    break;
	}
	for (piece = pawn; piece <= king; piece++){
	    side_attack_from_piece = side_attack & board->b[side][piece]; 
	    if (side_attack_from_piece){
		side_attack_from_sq = leadz (side_attack_from_piece);
		CLEARBIT (side_attack, side_attack_from_sq);
		if (xray[piece]){
		    AddXrayPiece (board, t, side_attack_from_sq, side, &side_attack, &xside_attack);
		}
		swaplist[n] = swaplist[n-1] + lastval;
		n++;
		lastval = -evals.Value[piece];
		break;
	    }
	}
    }
    
/****************************************************************************
 *
 *  At this stage, we have the swap scores in a list.  We just need to 
 *  mini-max the scores from the bottom up to the top of the list.
 *
 ****************************************************************************/
    --n;
    while (n){
	if (n & 1) {
	    if (swaplist[n] <= swaplist[n-1]){
		swaplist[n-1] = swaplist[n]; 
	    }
	} else {
	    if (swaplist[n] >= swaplist[n-1]){
		swaplist[n-1] = swaplist[n]; 
	    }
	}
	--n;
    }
    return (swaplist[0]);
}


int SwapOff(board_t *board, int move){
    return SwapOffEx (board, move, false);
}

/* This determines the is the SwapOff values is < threshold vs >= threshold.
   We take turns in playing captures until the side to move is above threshold
   or capturing is futile.
   We try to postpone expensive operations such a AttackToSqBitBoard as much as
   possible. Things could be further optimized by merging the implementation
   of AttackToSqBitBoard into this routine but that would yield unreadable
   code. I think.
*/
int SwapOffSignExNaive (board_t *board, int move, int threshold, int move_already_made){
    int f, t, piece, lastval;
    int side_attack_from_sq, xside_attack_from_sq;
    int side, xside; 
    int score;
    int from_sq;
    BitBoard side_attack, xside_attack, xside_attack_from_piece, side_attack_from_piece;

    f = FROMSQ (move);
    t = TOSQ (move);
    if(!move_already_made){
	side = ((board->friends[white] & bitboards.BitPosArray[f]) ? white : black);
    }else{
	side = ((board->friends[white] & bitboards.BitPosArray[t]) ? white : black);
    }
    xside = 1^side;

    side_attack = -1; /* hack */
    xside_attack =-1;

    from_sq=-1;
    if(!move_already_made){
	from_sq=f;
	if (move & PROMOTION){
	    score = evals.Value[PROMOTEPIECE (move)] - evals.ValueP;
	    lastval = evals.Value[PROMOTEPIECE (move)];
	} else {
	    score = (move & ENPASSANT ? evals.ValueP : evals.Value[board->cboard[t]]);
	    lastval = evals.Value[board->cboard[f]];
	}
    }else{
	lastval=evals.Value[board->cboard[t]];
	score=0; /* we cannot know what piece was there */
    }
    while (1){
	/* Them */
	if(score<threshold){
	    /* stand pat */
	    goto finish;
	}else if(score-lastval>=threshold){
	    /* futility */
	    score-=lastval;
	    goto finish;
	}
	if(xside_attack==-1){
	    xside_attack = AttackToSqBitBoard (board, t, xside);
	}
	if(from_sq>=0){
	    if(side_attack!=-1){
		CLEARBIT(side_attack, from_sq);
	    }
	    if(xray[board->cboard[from_sq]]){
		AddXrayPiece (board,t, from_sq, xside, &xside_attack,(side_attack!=-1)? &side_attack:NULL);
	    }
	}
	from_sq=-1;
	if (xside_attack == NULLBITBOARD){
	    goto finish;
	}
	for (piece = pawn; piece <= king; piece++) {
	    xside_attack_from_piece = xside_attack & board->b[xside][piece]; 
	    if (xside_attack_from_piece){
		xside_attack_from_sq = leadz (xside_attack_from_piece);
		from_sq=xside_attack_from_sq;
		score -= lastval;
		lastval = evals.Value[piece];
		break;
	    }
	}
	/* Us */
	if(score>=threshold){
	    /* stand pat */
	    goto finish;
	}else if(score+lastval<threshold){
	    /* futility */
	    score+=lastval;
	    goto finish;
	}
	if(side_attack==-1){
	    side_attack = AttackToSqBitBoard (board, t, side);
	    if(!move_already_made){
		CLEARBIT(side_attack,f);
		if(xray[board->cboard[f]]){
		    AddXrayPiece (board,t, f, side, &side_attack, &xside_attack);
		}
	    }
	}
	if(from_sq>=0){
	    CLEARBIT (xside_attack, from_sq);
	    if(xray[board->cboard[from_sq]]){
		AddXrayPiece (board,t, from_sq, side, &side_attack, &xside_attack);
	    }
	}
	from_sq=-1;
	if (side_attack == NULLBITBOARD){
	    goto finish;
	}
	for (piece = pawn; piece <= king; piece++){
	    side_attack_from_piece = side_attack & board->b[side][piece]; 
	    if (side_attack_from_piece){
		side_attack_from_sq = leadz (side_attack_from_piece);
		from_sq=side_attack_from_sq;
		score += lastval;
		lastval = evals.Value[piece];
		break;
	    }
	}
    }
 finish:
#ifdef DEBUG
    0; /* make gcc happy */
    int swap=0;
    swap=SwapOffEx(board, move, move_already_made);
    if(score<threshold){
	ASSERT(swap<threshold);
    }else{
	ASSERT(swap>=threshold);
    }
#endif    
    return score;
}


int SwapOffSignExClassic (board_t *board, int move, int threshold, int move_already_made){
    return SwapOffEx(board, move, move_already_made);
}

int SwapABWork(board_t *board, 
	       int from_sq, 
	       int to_sq, 
	       int lastval,
	       int side, 
	       int score,
	       int alpha,
	       int beta,
	       int ply,
	       BitBoard *side_attack_, 
	       BitBoard *xside_attack_){
    BitBoard side_attack_from_piece;
    int side_attack_from_sq;
    BitBoard side_attack, xside_attack;
    int xside;
    int piece;
    int lastval_save;
    xside=1^side;
    if(score>=beta){
	/* stand pat */
	goto finish;
    }else if(score+lastval<=alpha){
	/* futility */
	score+=lastval;
	goto finish;
    }
    alpha=MAX(score,alpha);  /* standing pat gives at least score */
    if(side_attack_==NULL){
	side_attack = AttackToSqBitBoard (board, to_sq, side);
	xside_attack = AttackToSqBitBoard (board, to_sq, xside);
    }else{
	side_attack = *side_attack_;
	xside_attack = *xside_attack_;
    }
    if(from_sq>=0){
	CLEARBIT(xside_attack, from_sq);
	if(xray[board->cboard[from_sq]]){
	    AddXrayPiece (board,to_sq, from_sq, side, &side_attack, &xside_attack);
	}
	from_sq=-1;
    }
    if ((side_attack) == NULLBITBOARD){
	goto finish;
    }
    for (piece = pawn; piece <= king; piece++){
	side_attack_from_piece = (side_attack) & board->b[side][piece]; 
	if (side_attack_from_piece){
	    side_attack_from_sq = leadz (side_attack_from_piece);
	    from_sq=side_attack_from_sq;
	    lastval_save=lastval;
	    lastval = evals.Value[piece];

	    break;
	}
    }
    if(side_attack_from_piece){
	int new_score;
	new_score=-SwapABWork(board, 
			      from_sq, 
			      to_sq, 
			      lastval,
			      xside, 
			      -score-lastval_save, /* score so far for xside */
			      -beta,
			      -alpha,
			      ply+1, 
			      &xside_attack, 
			      &side_attack);
	score=MAX(score, new_score);
    }
    

 finish:

    return score;
}

/* A recursive alpha-beta implementation. Cleaner code! */
int SwapOffSignExAB (board_t *board, int move, int threshold, int move_already_made){
    int piece, lastval;
    int side_attack_from_sq, xside_attack_from_sq;
    int side, xside; 
    int score;
    int from_sq, to_sq;
    int f;
    int ret;


    f=FROMSQ(move);
    to_sq = TOSQ (move);
    if(!move_already_made){
	side = ((board->friends[white] & bitboards.BitPosArray[f]) ? white : black);
    }else{
	side = ((board->friends[white] & bitboards.BitPosArray[to_sq]) ? white : black);
    }
    xside = 1^side;
    from_sq=-1;
    if(!move_already_made){
	from_sq=f;
	if (move & PROMOTION){
	    score = evals.Value[PROMOTEPIECE (move)] - evals.ValueP;
	    lastval = evals.Value[PROMOTEPIECE (move)];
	} else {
	    score = (move & ENPASSANT ? evals.ValueP : evals.Value[board->cboard[to_sq]]);
	    lastval = evals.Value[board->cboard[f]];
	}
    }else{
	lastval=evals.Value[board->cboard[to_sq]];
	score=0; /* we cannot know what piece was there */
    }
    
    ret= -SwapABWork(board, 
		       from_sq, 
		       to_sq, 
		       lastval,
		       xside, 
		       -score, /* score so far for xside */
		       -threshold,
		       -threshold+1,
		       1, /*ply */
		       NULL, 
		       NULL);

#ifdef DEBUG
    0;
    int swap=0;
    swap=SwapOffEx(board, move, move_already_made);
    if(ret<threshold){
	ASSERT(swap<=ret);
    }else{
	ASSERT(swap>=ret);
    }
#endif    
    return ret;

}



void AddXrayPiece (board_t *board, 
		   int to_sq, 
		   int x_sq, 
		   int side, 
		   BitBoard *side_attack, 
		   BitBoard *xside_attack)
/***************************************************************************
 *
 *  The purpose of this routine is to find a piece which attack through
 *  another piece (e.g. two rooks, Q+B, B+P, etc.)  Side is the side attacking
 *  the square where the swapping is to be done.
 *
 ***************************************************************************/
{
   int dir, first_occ_sq, piece;
   BitBoard pieces_on_ray;
   dir = directions[to_sq][x_sq];
   pieces_on_ray = bitboards.Ray[x_sq][dir] & board->blocker;
   if (pieces_on_ray == NULLBITBOARD)
      return;
   first_occ_sq = (to_sq < x_sq ? leadz (pieces_on_ray) : trailz (pieces_on_ray));
   piece = board->cboard[first_occ_sq];
   if ((piece == queen) || (piece == rook && dir > 3) || 
			    (piece == bishop && dir < 4)){
       if (bitboards.BitPosArray[first_occ_sq] & board->friends[side]){
	  if(side_attack){
	      *side_attack |= bitboards.BitPosArray[first_occ_sq];
	  }
       }else{
	  if(xside_attack){
	      *xside_attack |= bitboards.BitPosArray[first_occ_sq];
	  }
       }
   }
   return;
}
